#include <math.h>
#include "stdafx.h"
#include "Model.h"
#define WANT_STREAM                  // include.h will get stream fns
#define WANT_MATH                    // include.h will get math fns
#include "newmatap.h"                // need matrix applications
#include "newmatio.h"                // need matrix output routines
#include "MersenneTwister.h"
#include "specialfunctions2.h"
#include "cdpprior.h"
#include "cdpbase.h"
#include "cdpresult.h"
#include "cdp.h"
/*
#if defined(CDP_TBB)
	#include "cdpt.h"
#endif
*/
// Trying to implement the component swapping MH step
void CDP::ComponentSwap(CDPResult& result, MTRand& mt)
{
  vector<vector<int> > wk2d,wk2dOld;
  partition(result.W,result.K,wk2d);
  
  // 1. choose a component and a destination
  int j1,t1,j2,t2;
  //  findComponentsToSwap(result,wk2d,j1,t1,j2,t2,mt);
  j1=0;t1=2;j2=1;t2=1;

  // j1 t1 is the original location, j2 t2 is the swapped location.
  if(j1==j2 || wk2d[j2*result.T+t2].size()!=0)
    {
      //      fprintf(stdout, "Nothing to swap now\n");
      return; // nothing can be swapped now
    }
  else
    {
      // 2. set new w and new k
      int *Wnew;
      Wnew = new int[result.N];
      int *Knew;
      Knew = new int[result.N];
      for(int i=0;i<result.N;i++)
	{
	  Wnew[i] = result.W[i];
	  Knew[i] = result.K[i];
	}
      vector<int>::iterator it;
      for(it=wk2d[j1*result.T+t1].begin();it!=wk2d[j1*result.T+t1].end();it++)
	{
	  Wnew[*it]=j2;Knew[*it]=t2; // move what was in j1,t1 to j2,t2
	}
      for(it=wk2d[j2*result.T+t2].begin();it!=wk2d[j2*result.T+t2].end();it++)
	{
	  Wnew[*it]=j1;Knew[*it]=t1; // move what was in j2,t2 to j1,t1
	}
      
      // 3.  keep wk2d up to date too
      wk2dOld = wk2d;
      vector<int> foo;
      foo = wk2d[j1*result.T+t1];
      wk2d[j1*result.T+t1] = wk2d[j2*result.T+t2]; // move the j2,t2 obs to j1,t1 (an empty list)
      wk2d[j2*result.T+t2] = foo; // the j1,t1 obs to j2,t2 (a nonempty list)
      
      // 4. set mean and covariance of newly occupied component
      RowVector newmu2 = result.mu[j1*result.T+t1]; 
      SymmetricMatrix newSigma2 = result.Sigma[j1*result.T+t1];
      LowerTriangularMatrix newSigma2_i = result.L_i[j1*result.T+t1];
      double newSigma2_logdet = result.Sigma_log_det[j1*result.T+t1];
      
      // 5. set mean and covariance of newly vacated component
      SymmetricMatrix blah = result.Phi[j1]*prior.nu;
      blah = blah.i();
      LowerTriangularMatrix ltmat = Cholesky(blah);
      SymmetricMatrix newSigma1 = msf.invwishartrand(prior.nu+2,ltmat,mt);   //sampled from its prior
      
      ltmat = Cholesky(newSigma1);
      LowerTriangularMatrix newSigma1_i = ltmat.i();
      double newSigma1_logdet = msf.logdet(ltmat);
      
      blah = newSigma1*prior.gamma;
      ltmat = Cholesky(blah);
      RowVector newmu1 = msf.mvnormrand(result.m[j1],ltmat,mt);
            
      // 6. set new cluster weights
      RowVector newq;
      RowVector newqV;
      sampleP(Wnew,result.N,result.alpha0,result.J,newq,newqV,mt);

      // 7. set new component weights for the cluster containing the newly occupied component
      vector<int> KJ2new;
      KJ2new.clear();
      for(int t=0;t<result.T;t++)
	for(unsigned int i=0;i<wk2d[j2*result.T+t].size();i++)
	  KJ2new.push_back(t);
      
      RowVector newp2;
      RowVector newpV2;
      sampleP(KJ2new,KJ2new.size(),result.alpha[j2],result.T,newp2,newpV2,mt);
      
      // 8 set new component weights for the cluster containing the newly vacated component
      vector<int> KJ1new;
      KJ1new.clear();
      for(int t=0;t<result.T;t++)
	for(unsigned int i=0;i<wk2d[j1*result.T+t].size();i++)
	  KJ1new.push_back(t);

      RowVector newp1;
      RowVector newpV1;
      sampleP(KJ1new,KJ1new.size(),result.alpha[j1],result.T,newp1,newpV1,mt);
      
      // 9. evaluate acceptance probability:
//       cout << "newpq = " << newq << "\n";
//       cout << "newp2 = " << newp2 << "\n";
//       cout << "newp1 = " << newp1 << "\n";

      double n1 = piece_1(wk2d,newq,newp2,j2,newp1,j1) ;
      double d1 = piece_1(wk2dOld,result.q,result.p[j2],j2,result.p[j1],j1);

      // cancels with the same piece in the transition probability
      //      double n2 = piece_2(newmu1,newSigma1, newSigma1_T_i, newSigma1_logdet,result.m[j1],result.Phi[j1],result.Phi_log_det[j1], prior.nu,prior.gamma);

      double d2 = piece_2(result.mu[j1*result.T+t1],result.Sigma[j1*result.T+t1],result.L_i[j1*result.T+t1],result.Sigma_log_det[j1*result.T+t1], result.m[j1],result.Phi[j1],result.Phi_log_det[j1],prior.nu,prior.gamma);

      double n3 = piece_2(newmu2,newSigma2, newSigma2_i, newSigma2_logdet,result.m[j2],result.Phi[j2],result.Phi_log_det[j2], prior.nu,prior.gamma);

      // cancels with the same piece in the transition probability
      //      double d3 = piece_2(result.mu[j2*result.T+t2],result.Sigma[j2*result.T+t2],result.Sigma_T_i[j2*result.T+t2],result.Sigma_log_det[j2*result.T+t2],result.m[j2],result.Phi[j2],result.Phi_log_det[j2],prior.nu,prior.gamma);

      double n4 = piece_3(newqV,result.alpha0);
      double d4 = piece_3(result.qV,result.alpha0);
      double n5 = piece_3(newpV1,result.alpha[j1]);
      double d5 = piece_3(result.pV[j1],result.alpha[j1]);
      double n6 = piece_3(newpV2,result.alpha[j2]);
      double d6 = piece_3(result.pV[j2],result.alpha[j2]);
      
      double* mya = new double[result.T];
      for(int t=0;t<result.T;t++)
	mya[t] = (double) wk2dOld[j2*result.T+t].size();
      
      double n7 = piece_4(mya,result.alpha[j2],result.pV[j2]);      
      
      for(int t=0;t<result.T;t++)
	mya[t] = (double) wk2dOld[j1*result.T+t].size();
      double n8 = piece_4(mya,result.alpha[j1],result.pV[j1]);

      for(int t=0;t<result.T;t++)
	mya[t] = (double) wk2d[j1*result.T+t].size();
      double d8 = piece_4(mya,result.alpha[j1],newpV1);
      
      for(int t=0;t<result.T;t++)
	mya[t] = (double) wk2d[j2*result.T+t].size();
      double d7 = piece_4(mya,result.alpha[j2],newpV2);

      double* myb = new double[result.J];
      for(int j=0;j<result.J;j++)
	{
	  myb[j]=0;
	  for(int t=0;t<result.T;t++)
	    myb[j]+=wk2dOld[j*result.T+t].size();
	}
      double n9 = piece_4(myb,result.alpha0,result.qV);
      
      for(int j=0;j<result.J;j++)
	{
	  myb[j]=0;
	  for(int t=0;t<result.T;t++)
	    myb[j]+=wk2d[j*result.T+t].size();
	}
      double d9 = piece_4(myb,result.alpha0,newqV);

      
      double logaccprob = n1 + n3 + n4 + n5 + n6 +n7 + n8 +n9 - d1 -d2 -d4 -d5 -d6 - d7 -d8 -d9;
      logaccprob = exp(logaccprob);
      logaccprob = logaccprob>1 ? 1 : logaccprob;
      
//       if(isnan(logaccprob))
// 	{
// 	  cout << "\n" << n1 << " " << n3 << " " << n4 << " " << n5 << " " << n6 << " " << n7 << " " << n8 << " " << n9 << "\n";
// 	  cout << d1 << " " << d2  << " " << d4 << " " << d5 << " " << d6 << " " << d7 << " " << d8 << " " << d9 << "\n";
// 	  exit(1);
// 	}


      if(mt()<logaccprob)
	{      
	  fprintf(stdout, "Swapping (%d, %d) with (%d, %d) ...",j1,t1,j2,t2);
	  cout << "acceptance prob = " << logaccprob << "\n";
	  //	  cout << "accepted!\n";
	  
	  // copy the new values to the result object
	  for(int i=0;i<result.N;i++)
	    {
	      result.W[i] = Wnew[i];
	      result.K[i] = Knew[i];
	    }
	  result.mu[j1*result.T+t1] = newmu1;
	  result.Sigma[j1*result.T+t1] = newSigma1;
	  result.Sigma_log_det[j1*result.T+t1] = newSigma1_logdet;
	  result.L_i[j1*result.T+t1] = newSigma1_i;
	  
	  result.mu[j2*result.T+t2] = newmu2;
	  result.Sigma[j2*result.T+t2] = newSigma2;
	  result.Sigma_log_det[j2*result.T+t2] = newSigma2_logdet;
	  result.L_i[j2*result.T+t2] = newSigma2_i;
	  
	  result.q = newq;
	  result.qV = newqV;
	  result.p[j1] = newp1;
	  result.pV[j1] = newpV1;
	  result.p[j2] = newp2;
	  result.pV[j2] = newpV2;
	}
      else
	{

	//	cout << "rejected\n";
	}
      
      // 10. delete stuff
      delete [] mya;
      mya = NULL;
      delete [] myb;
      myb = NULL;
      for (int i=0; i<result.N; ++i)
	delete Wnew[i];
      delete [] Wnew;
      Wnew = NULL;
      for (int i=0; i<result.N; ++i)
	delete Knew[i];
      delete [] Knew;
      Knew = NULL;
      return;
    }
}

double CDP::piece_3(RowVector& V, double alpha)
{
  int dim = V.Ncols();
  double d=0;
  for(int i=0;i<dim;i++)
    {
      if(V[i]==1)
	break;
      else
	d+=msf.betapdf(V[i],1,alpha,1);
    }
  return d;
}

double CDP::piece_4(double* a, double alpha, RowVector& V) {
  int T = V.Ncols();
  
  double* b = new double[T];
  int i;
  for (i = 0; i< T; i++) {
     b[i] = 0;
  }

  b[T-1] = 0;
  for (i = T-2; i >=0;i--) {
    b[i] = b[i+1] + a[i+1];
  }       

  double logp = 0;
  for (i = 0; i< T; i++) {
    if (V[i]==1) 
      break;
    else
      logp+= msf.betapdf(V[i],1+a[i],alpha + b[i],1);
  }
  delete [] b;
  b = NULL;
  return logp;
}


// sums the log probabilities associated with the indicator variables being drawn from the associated discrete
// probability distributions -- ie 
// w_i ~ \sum_j=1^J q_j delta_j()
// k_i | w_i ~ \sum_k=1^K p_{w_i,k} delta_k()
double CDP::piece_1(vector<vector<int> >& wk2d,RowVector& q,RowVector& p2, int j2, RowVector& p1, int j1)
{
  double logprob = 0;
  for(int j=0;j<prior.J;j++)
    {
      if(j==j2)
	{
	  for(int t=0;t<prior.T;t++)
	    {
	      logprob+=wk2d[j*prior.T+t].size()*log(p2[t]);
	      logprob+=wk2d[j*prior.T+t].size()*log(q[j]);
	    }
	}
      else if(j==j1)
	{
	  for(int t=0;t<prior.T;t++)
	    {
	      logprob+=wk2d[j*prior.T+t].size()*log(p1[t]);
	      logprob+=wk2d[j*prior.T+t].size()*log(q[j]);
	    }
	}
      else
	{
	  for(int t=0;t<prior.T;t++)
	    {
	      logprob+=wk2d[j*prior.T+t].size()*log(q[j]);
	    }
	}
    }
  return logprob;
}

// computed the log of tge NIW density for (newmu, newSigma) ~ NIW(m,Phi,nu,gamma)
double CDP::piece_2(RowVector& newmu, SymmetricMatrix& newSigma, LowerTriangularMatrix& newSigmaUTI, double newSigmaLogDet, RowVector& m, SymmetricMatrix& Phi, double philogdet, double nu, double gamma)
{
  double logd = 0;
  double logd2 = 0;
  SymmetricMatrix temp = newSigma*gamma;
  LowerTriangularMatrix temp2 = Cholesky(temp);
  double templogdet = msf.logdet(temp2);
  UpperTriangularMatrix temp3 = temp2.t().i();

  // MVN(mu | m, gamma*Sigma)
  logd = msf.mvnormpdf(newmu.Store(),m.Store(),temp3,temp3.Ncols(),1,templogdet);

  temp = Phi*nu;
  // IW(Sigma | nu+2, nu*Phi)
  logd2= msf.invwishartpdf(newSigmaUTI,newSigmaLogDet,nu+2,temp,philogdet,temp3.Ncols(),1);
  //  cout << newSigma << "\n\n" << Phi << "\n\n" << nu << " == " << logd2 << "\n";
  
  return logd+logd2;
}

void CDP::findComponentsToSwap(CDPResult& result, vector<vector<int> >& wk2d, int& j1, int& t1, int& j2, int& t2,MTRand& mt)
{
//   for(unsigned int a=0;a<wk2d.size();a++)
//     {
//       cout << a << "  " << wk2d[a].size() << "\n";
//     }

  double* clustweights;
  clustweights = new double[result.J];
  for(int j=0;j<result.J;clustweights[j++]=1.)
    ;
  
  double* compweights;
  compweights = new double[result.T];
  for(int t=0;t<result.T;compweights[t++]=1.)
    ;
  
  int ITERTRY=10;
  int ntries=0;
  do
    {
      j1 = sample(clustweights,result.J,mt);
      t1 = sample(compweights,result.T,mt);
      //      cout << "sampled " << j1 << " " << t1 << "\n";

      // choose a new cluster
      do
	{
	  j2 = sample(clustweights,result.J,mt);
	  //	  cout << "sampled " << j2 << "\n";
	} 
      while(j2==j1);  
      
      // see if it has an empty component
      for(t2=0;t2<result.T;t2++)
	if(wk2d[j2*result.T+t2].size()==0)
	  {
	    break;
	  }
    }
  while(t2>=result.T && ntries++<=ITERTRY);

  delete [] clustweights;
  clustweights = NULL;
  delete [] compweights;
  compweights = NULL;

  return;
}


