%module cdp
%{
#define SWIG_FILE_WITH_INIT
#include "cdpcluster.h"
#include "cdpemcluster.h"
#include "mvnpdf.h"

#include <stdexcept>
%}
%include numpy.i
%include mvnpdf.i

%init %{
import_array();
%}

%exception getdevice {
   try {
      $action
   } catch (runtime_error &e) {
      PyErr_SetString(PyExc_RuntimeError, const_cast<char*>(e.what()));
      return NULL;
   }
}

%exception setdevice {
   try {
      $action
   } catch (runtime_error &e) {
      PyErr_SetString(PyExc_RuntimeError, const_cast<char*>(e.what()));
      return NULL;
   }
}

%exception getgpuchunksize {
   try {
      $action
   } catch (runtime_error &e) {
      PyErr_SetString(PyExc_RuntimeError, const_cast<char*>(e.what()));
      return NULL;
   }
}

%exception setgpuchunksize {
   try {
      $action
   } catch (runtime_error &e) {
      PyErr_SetString(PyExc_RuntimeError, const_cast<char*>(e.what()));
      return NULL;
   }
}
class cdpcluster {
public:
%apply (int DIM1, int DIM2, double* IN_ARRAY2) {(int n, int d, double* x)}
	cdpcluster(int n, int d, double* x);
	~cdpcluster(void);
	void makeResult();
	void run();
	void step();
	void stepburn();
	void setseed(int x);
	void setT(int t);
	void setJ(int j);
	void setBurnin(int t);
    void setIter(int t);
    void setVerbose(bool verbosity);
	int getBurnin();
    int getIter();
	int getn();
	int getd();
	int getclustN();
	int getT();
	int getJ();
	//int getK(int idx);
%apply (int DIM1, int* ARGOUT_ARRAY1) {(int d, int* res)}
	void getK(int d, int* res);
	void setlambda0( double lambda0 );
	double getlambda0();
	void setphi0( double phi0 );
	double getphi0();
	void setnu0( double nu0 );
	double getnu0();
	void setgamma( double gamma );
	double getgamma();
	void setnu( double nu );
	double getnu();
	void sete0( double e0 );
	double gete0();
	void setf0( double f0 );
	double getf0();
	void setee( double ee );
	double getee();
	void setaa( double aa );
	double getaa();
	void setff( double ff );
	double getff();	
	
	double getMu(int idx, int pos);
	double getm(int idx, int pos);
	double getSigma(int i, int j, int k);
	double getPhi(int i, int j, int k);
	double getp(int idx);
	
	void setgpunchunksize(int x);
	int getgpunchunksize();
	void setdevice(int x);
	int getdevice();
	int getnumberdevices();
	void setnumberdevices(int x);
	
	bool samplem();
	void samplem(bool x);
	bool samplePhi();
	void samplePhi(bool x);
	bool samplew();
	void samplew(bool x);
	bool sampleq();
	void sampleq(bool x);
	bool samplealpha0();
	void samplealpha0(bool x);
	bool samplemu();
	void samplemu(bool x);
	bool sampleSigma();
	void sampleSigma(bool x);
	bool samplek();
	void samplek(bool x);
	bool samplep();
	void samplep(bool x);
	bool samplealpha();
	void samplealpha(bool x);
	void loadMu(int n, int d, double* x);
	void loadm(int n, int d, double* x);
	void loadp(int n, int d, double* x);
	void loadpV(int n, int d, double* x);
%apply (int DIM1, int DIM2, int DIM3, double* IN_ARRAY3) {(int i, int j, int k, double* x)}
	void loadSigma(int i, int j, int k , double* x);
	void loadPhi(int i, int j, int k , double* x);	
%apply (int DIM1, double* IN_ARRAY1) {(int i, double* x)}
	void loadW(int i, double* x);
	void loadK(int i, double* x);
	void loadq(int i, double* x);
	void loadqV(int i, double* x);
	void loadalpha(int i, double* x);
%apply (int DIM1, double* IN_ARRAY1) {(int n, double* x)}
	void loadref(int n, double* x); 


	void printModel();

	private:
	void loadRows(double* from, int* to, int cols);
	void loadRowsCols(double* from, vector<RowVector>& to, int rows, int cols);
	void loadRowsCols(double* from, vector<RowVector>& to, int rows, int cols);
	void loadRows(double* from, RowVector& A, int cols);
	CDPResult* param;
	//CDPResult result;
	Model model;
	CDP cdp;
	MTRand mt;
};

class cdpemcluster {
	public:
	~cdpemcluster(void); 
%apply (int DIM1, int DIM2, double* IN_ARRAY2) {(int n, int d, double* x)}
	cdpemcluster(int n, int d, double* x);

	//void makeResult();
	void run();
	//void step();
	//void stepburn();
	void setseed(int x);
	void setT(int t);
	void setJ(int j);
	void setBurnin(int t);
    void setIter(int t);
    void setVerbose(bool verbosity);
	int getBurnin();
    int getIter();
	int getn();
	int getd();
	int getclustN();
	int getT();
	int getJ();
	//int getK(int idx);
%apply (int DIM1, int* ARGOUT_ARRAY1) {(int d, int* res)}
	void getK(int d, int* res);
	void setlambda0( double lambda0 );
	double getlambda0();
	void setphi0( double phi0 );
	double getphi0();
	void setnu0( double nu0 );
	double getnu0();
	void setgamma( double gamma );
	double getgamma();
	void setnu( double nu );
	double getnu();
	void sete0( double e0 );
	double gete0();
	void setf0( double f0 );
	double getf0();
	void setee( double ee );
	double getee();
	void setaa( double aa );
	double getaa();
	void setff( double ff );
	double getff();	
	
	double getMu(int idx, int pos);
	double getm(int idx, int pos);
	double getSigma(int i, int j, int k);
	double getPhi(int i, int j, int k);
	double getp(int idx);
	
	void setgpunchunksize(int x);
	int getgpunchunksize();
	void setdevice(int x);
	int getdevice();
	int getnumberdevices();
	void setnumberdevices(int x);
	
	bool samplem();
	void samplem(bool x);
	bool samplePhi();
	void samplePhi(bool x);
	bool samplew();
	void samplew(bool x);
	bool sampleq();
	void sampleq(bool x);
	bool samplealpha0();
	void samplealpha0(bool x);
	bool samplemu();
	void samplemu(bool x);
	bool sampleSigma();
	void sampleSigma(bool x);
	bool samplek();
	void samplek(bool x);
	bool samplep();
	void samplep(bool x);
	bool samplealpha();
	void samplealpha(bool x);
	void loadMu(int n, int d, double* x);
	void loadm(int n, int d, double* x);
	void loadp(int n, int d, double* x);
	void loadpV(int n, int d, double* x);
%apply (int DIM1, int DIM2, int DIM3, double* IN_ARRAY3) {(int i, int j, int k, double* x)}
	void loadSigma(int i, int j, int k , double* x);
	void loadPhi(int i, int j, int k , double* x);	
%apply (int DIM1, double* IN_ARRAY1) {(int i, double* x)}
	void loadW(int i, double* x);
	void loadK(int i, double* x);
	void loadq(int i, double* x);
	void loadqV(int i, double* x);
	void loadalpha(int i, double* x);


	void printModel();

	private:
	void loadRows(double* from, int* to, int cols);
	void loadRowsCols(double* from, vector<RowVector>& to, int rows, int cols);
	void loadRowsCols(double* from, vector<RowVector>& to, int rows, int cols);
	void loadRows(double* from, RowVector& A, int cols);
	
	CDP cdpfunctions;
	CDPResultEM* param;
	
	Model model;
	CDP_EM cdp;
	MTRand mt;
	bool verbose;
};
